// Description: adds highlighting to the nodes below it
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> the mode options are points, lines, normals, bbox, and fill
//</ul>
//
// Category: attributes
// Author: Alex Hill
// Revision: 11/01/01
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfGeode.h>
#include <ygNetMessages.h>
#include "highlight.h"

extern "C" ygNode* construct_highlight(const char* name,bool master) { return new highlight(name,master); }

highlight::highlight(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("highlight",true);
    highlight_ = new pfHighlight;
    setActive(true);
    setMode("lines");
    color.set(1,1,1,1);
    setColor(color);
    length = 1.0;
    width = 1.0;
    highlight_->setNormalLength(0.0,length);
    highlight_->setPntSize(width);
    highlight_->setLineWidth(width);
    addNetMessage("off",&active,YG_NET_BOOL_NEGATE);
    addNetMessage("color",&color,YG_NET_VEC4);
    addNetMessage("mode",&mode,YG_NET_STRING);
    addNetMessage("length",&length,YG_NET_FLOAT);
    addNetMessage("width",&width,YG_NET_FLOAT);
    }

highlight::~highlight(void)
    {
    }


void highlight::reset(void)
    {
    //set active to true
    resetNetMessage("off");
    setActive(active);
    //set the highlight mode to lines
    resetNetMessage("mode");
    setMode(mode);
    //set the highlight color to white
    resetNetMessage("color");
    setColor(color);
    //set the highlight normal length to 1.0
    resetNetMessage("width");
    setWidth(width);
    //set the highlight width to 1.0
    resetNetMessage("length");
    setLength(length);
    ygNode::reset();
    }


void highlight::message(const ygMessage& msg)
    {
    //include any combination of options
    if (msg == "mode")
        {
        setMode(mode);
        }
    //set the color with or without alpha
    else if (msg == "color")
        {
        setColor(color);
        }
    //set the line width and point size
    else if (msg == "width")
        {
        setWidth(width);
        }
    //set the length of normals
    else if (msg == "length")
        {
        setLength(length);
        }
    //turn highlighting on
    else if (msg == "on")
        {
        setActive(true);
        netMessageChanged("off");
        }
    //turn highlighting off
    else if (msg == "off")
        {
        setActive(active);
        }
    else
        ygNode::message(msg);
    }

void highlight::app(void)
    {
    if (isChanged(YG_CHANGE_GEOMETRY))
        setActive(active);
    ygNode::app();
    }

void highlight::setGSetHlight(pfHighlight* hlight,pfNode* node)
    {
    int i;
    //if node type is pfGeode then set geoset properties
    if (node->isOfType(pfGeode::getClassType()))
        {
        pfGeode *geode = (pfGeode *)node;
        //for each geode, call setGeosetTexture
        for (i=0; i < geode->getNumGSets(); i++)
            geode->getGSet(i)->setHlight(hlight);
        }
    //else, if node type is pfGroup
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup *group = (pfGroup *)node;
        //call this function recursively on all children
        for (i=0; i < group->getNumChildren(); i++)
            setGSetHlight(hlight,group->getChild(i));
        }
    }

void highlight::setActive(bool val)
    {
    active = val;
    if (active)
        setGSetHlight(highlight_,pfnode());
    else
        setGSetHlight(NULL,pfnode());
    }

void highlight::setMode(const ygString& modes)
    {
    uint mode = 0;
    ygString* token;
    int i=0;
    while (token = modes.nextToken(" ",&i))
        {
        if (*token == "points")
            mode |= PFHL_POINTS;
        else if (*token == "lines")
            mode |= PFHL_LINES;
        else if (*token == "normals")
            mode |= PFHL_NORMALS;
        else if (*token == "bbox")
            mode |= PFHL_BBOX_LINES;
        else if (*token == "fill")
            mode |= PFHL_FILL;
        }
    highlight_->setMode(mode);
    }

void highlight::setColor(const pfVec4& val)
    {
    color = val;
    highlight_->setColor(PFHL_FGCOLOR,color[0],color[1],color[2]);
    highlight_->setAlpha(color[3]);
    }

void highlight::setWidth(float val)
    {
    width = val;
    highlight_->setPntSize(val);
    highlight_->setLineWidth(val);
    }

void highlight::setLength(float val)
    {
    length = val;
    highlight_->setNormalLength(0.0,val);
    }

