// Description: adjust the texture matrix of all nodes below this node
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> rotations are applied before translation
//<li> use the startPosition message to do a pre-translation if necessary
//<li> the texture matrix of all instanced object nodes will be changed
//<li> use the cache option on child objects to avoid unwanted side effects
//</ul>
//
// Category: Attributes
// Author: Alex Hill
//           11/01/01
// Revision: 12/05/04 Alex Hill - fixed matrix key for client nodes
//
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfGroup.h>
#include <Performer/pfdu.h>
#include "ygNetMessages.h"
#include "textureTransform.h"

extern "C" ygNode* construct_textureTransform(const char* name,bool master) { return new textureTransform(name,master); }


textureTransform::textureTransform(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("textureTransform",true);
    gstate = NULL;
    startPosition.set(0,0);
    position.set(0,0);
    orientation.set(0,0);
    size.set(1,1);
    matrix = (pfMatrix*) pfMalloc(sizeof(pfMatrix), pfGetSharedArena());
    matrix->makeIdent();
    addNetMessage("startPosition",&startPosition,YG_NET_VEC2);
    dontSendNetMessage("startPosition");
    addNetMessage("position",&startPosition,YG_NET_VEC2);
    dontSendNetMessage("position");
    addNetMessage("orientation",&orientation,YG_NET_VEC2);
    dontSendNetMessage("orientation");
    addNetMessage("size",&size,YG_NET_VEC2);
    dontSendNetMessage("size");
    addNetMessage("matrix",matrix,YG_NET_MATRIX);
    //dont include in documentation
    dontSaveNetMessage("matrix");
    }

textureTransform::~textureTransform(void)
    {
    }

void textureTransform::reset(void)
    {
    //reset texture matrix to identity
    matrix->makeIdent();
    startPosition.set(0,0);
    position.set(0,0);
    orientation.set(0,0);
    size.set(1,1);
    setMatrix();
    ygNode::reset();
    } 

void textureTransform::setTexMat(pfNode *node)
    {
    int i;
    //if node type is pfGeode then set geoset material
    if (node->isOfType(pfGeode::getClassType()))
        {
        pfGeode *geode = (pfGeode *)node;
        for (i=0; i < geode->getNumGSets(); i++)
            setGeosetTexMat(geode->getGSet(i));
        }
    //else, if node type is pfGroup
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup *group = (pfGroup *)node;
        //call setMaterial recursively on all children
        for (i=0; i < group->getNumChildren(); i++)
            setTexMat(group->getChild(i));
        }
    }

void textureTransform::setGeosetTexMat(pfGeoSet *gset)
    {
    //find and setup geostate texture matrix
    gstate = gset->getGState();
    gstate->setMode(PFSTATE_ENTEXMAT, PF_ON);
    gstate->setAttr(PFSTATE_TEXMAT, matrix);
    }


void textureTransform::message(const ygMessage& msg)
    {
    //set pre-rotation translational component
    if (msg == "startPosition")
        {
        setMatrix();
        }
    //set rotational component
    else if (msg == "orientation")
        {
        setMatrix();
        }
    //set translational component
    else if (msg == "position")
        {
        setMatrix();
        }
    //set scaling component
    else if (msg == "size")
        {
        if (msg.numArgs() == 1)
            size[1] = size[0];
        setMatrix();
        }
    //receive matrix key
    else if (msg == "matrix")
        {
        return;
        }
    else
        ygNode::message(msg);
    }

void textureTransform::app(void)
    {
    //if any child has changed then setMatrix for all child nodes
    if (isChanged(YG_CHANGE_GEOMETRY))
        setTexMat(pfnode());
    ygNode::app();
    }

void textureTransform::setMatrix(void)
    {
    //create scaling matrix
    matrix->makeScale(size[0], size[1], 1);
    pfMatrix tmpMatrix;
    //multiply by pre-rotational translation
    tmpMatrix.makeTrans(startPosition[0], startPosition[1], 0);
    matrix->postMult(tmpMatrix);
    //multiply by rotation
    tmpMatrix.makeEuler(orientation[0], orientation[1], 0);
    matrix->postMult(tmpMatrix);
    //multiply by translation
    tmpMatrix.makeTrans(position[0], position[1], 0);
    //set texture matrix
    matrix->postMult(tmpMatrix);
    netMessageChanged("matrix");
    }

