// Description: sets up an independent Performer switch state for each user
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> see reset method for default settings
//<li> creates a switch using a Performer pfSwitch
//<li> this node does not stop YG application processing from occuring
//<li> this node is frequently used to toggle geometry on and off
//</ul>
//
// Category: Selection
// Author: Alex Hill
// Revision: 10/01/02
//
#include <ygWorld.h>
#include <ygNodeDB.h>
#include <ygNetMessages.h>
#include <ygCAVEViewer.h>
#include "userSwitch.h"

extern "C" ygNode* construct_userSwitch(const char* name,bool master) { return new userSwitch(name,master); }

userSwitch::userSwitch(const char* name,bool master) : ygNode(name,master,false)
    {
    setClassName("userSwitch",true);
    //create pfSwitch for the pfnode
    sw = new pfSwitch;
    sw->setVal(PFSWITCH_ON);
    setPfNode(sw);
    addNetMessage("on",&users,YG_NET_STRING_DELETE);
    dontSendNetMessage("on");
    dontSaveNetMessage("on");
    addNetMessage("off",&users,YG_NET_STRING_ADD);
    dontSendNetMessage("off");
    }

userSwitch::~userSwitch(void)
    {
    }

void userSwitch::reset(void)
    {
    sw->setVal(PFSWITCH_ON);
    resetNetMessage("users");
    ygNode::reset();
    }

void userSwitch::message(const ygMessage& msg)
    {
    //set the switch to on for all or a single user
    if (msg == "on")
        {
        if (msg.numArgs() == 0 && isNetMaster())
            users.clear();
        setState(users);
        netMessageChanged("off");
        }
    //set the switch to off for all or a single user
    else if (msg == "off")
        {
        if (msg.numArgs() == 0 && isNetMaster())
            {
            users.clear();
            //for each user in the world
            for (int i=0; i < ygWorld::world()->numUsers(); i++)
                {
                ygNode* node = (ygNode*)ygWorld::world()->user(i);
                users += node->name();
                users += " ";
                }
            }
        setState(users);
        netMessageChanged("off");
        }
    //toggle the state of the switch for all or a single user
    else if (msg == "toggle")
        {
        if (msg.numArgs() < 2)
            {
            if (msg.numArgs() == 1)
                toggle(msg.stringArg(0));
            else
                {
                //for each user in the world
                for (int i=0; i < ygWorld::world()->numUsers(); i++)
                    {
                    ygNode* node = (ygNode*)ygWorld::world()->user(i);
                    toggle(node->name());
                    }
                setState(users);
                netMessageChanged("off");
                }
            }
        else
            msg.error(name(),"(wrong number of arguments)");
        }
    else
        ygNode::message(msg);
    }

bool userSwitch::on(const ygString& user)
    {
    //remove this user from the users list
    ygString *curUser;
    int i=0;
    bool removed = false;
    ygString newUsers;
    while (curUser = users.nextToken(" ",&i))
            {
            if (*curUser == user)
                removed = true;
            else
                {
                newUsers += *curUser;
                newUsers += " ";
                }
            delete curUser;
            }
    if (removed)
        users = newUsers;
    return removed;
    }

bool userSwitch::off(const ygString& user)
    {
    //add this user to the users list
    ygString *curUser;
    int i=0;
    bool added = true;
    ygString newUsers;
    while (curUser = users.nextToken(" ",&i))
        {
        if (*curUser == user)
            added = false;
        delete curUser;
        }
    if (added)
        {
        users += user;
        users += " ";
        }
    return added;
    }

void userSwitch::toggle(const ygString& user)
    {
    if (!off(user))
        on(user);
    }

void userSwitch::setState(const ygString& users)
    {
    ygString *user;
    ygNode* node;
    ygCAVEViewer* viewer;
    viewer = (ygCAVEViewer*)ygWorld::world()->viewer();
    int i=0;
    float state = PFSWITCH_ON;
    while (user = users.nextToken(" ",&i))
            {
            node = ygNodeDB::find(*user);
            if (node && (node->isNetMaster() || viewer->isCloned(node)))
                state = PFSWITCH_OFF;
            delete user;
            }
    sw->setVal(state);
    }
