// Revision: 05/02/03 Alex Hill - added --viewer option for distributed applications

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <Performer/pf/pfChannel.h>
#include <Performer/pfdu.h>
#include <Performer/pfutil.h>
#include <pfcave.h>
#include <QUANTA/QUANTA.hxx>
#include "ygNet.h"
#include "ygParser.h"
#include "ygWorld.h"
#include "ygNodeDB.h"
#include "ygNet.h"
#include "ygNode.h"
#include "ygUtil.h"
#include "ygControl.h"
#include "ygUser.h"
#include "ygCAVEViewer.h"
#include "ygFileRequest.h"

using namespace std;

static void initGraphics(pfChannel *,void *);
static void checkNetMessages(void);
static void checkStdinMessages(void);
static void handleMessage(char *message);
static void addCAVEClasses(void);

static bool createView=true;
static bool quitFlag=false;


void printUsers(ygWorld *world)
    {
    int i,j;
    printf("\n%d users:\n",world->numUsers());
    for (i=0; i < world->numUsers(); i++)
        {
        ygUser *u = world->user(i);
        printf("  %s; ",u->name().c_str());
        printf("head=%p",u->head());
        if (u->head())
            printf("(%s)",u->head()->name().c_str());
        printf("; wands=");
        for (j=0; j < u->numWands(); j++)
            {
            printf(" %p",u->wand(j));
            if (u->wand(j))
                printf("(%s)",u->wand(j)->name().c_str());
            }
        printf("\n");
        }
    }


static void parseCommandLine(int argc,char **argv,ygWorld * world)
    {
    ygParser parser;
    int i;
    for (i=1; i < argc; i++)
        {
        if (!strncmp(argv[i],"--",2))
            {
            if (!strcasecmp(argv[i],"--noview"))
                createView = false;
            else if (!strcasecmp(argv[i],"--viewer"))
                ((ygCAVEViewer*)world->viewer())->setUser(argv[++i]);
            else
                fprintf(stderr,"WARNING: unknown command line "
                    "option '%s'\n",argv[i]);
            }
        else
            {
            parser.parseFile(argv[i],world->localRoot());
            }
        }
    }


int main(int argc,char **argv)
    {
    ygWorld *world;
    ygNode *root;
    setlinebuf(stdin);
    ygPrePFInit();
    pfInit();
    pfCAVEConfig(&argc,argv,NULL);
    CAVESetOption(CAVE_PROJ_INCLUDENAVIGATION,1);
    ygPrePFConfig();
    addCAVEClasses();
    pfConfig();
    pfCAVEInitChannels();
    world = new ygWorld;
    root = ygNodeDB::find("root");
    if (root)
        {
        printf("Found existing root node\n");
        world->setRoot(root);
        }
    else
        {
        printf("Creating new root node\n");
        world->createRoot();
        }
    parseCommandLine(argc,argv,world);
    if (world->viewer())
        world->viewer()->reset();
    pfCAVEMasterChan()->setScene(world->pfscene());
    if (createView)
        {
        pfCAVEPreDrawFunc((pfChanFuncType)initGraphics);
        while ((!CAVEgetbutton(CAVE_ESCKEY)) && (!quitFlag) &&(!world->quit()))
            {
            pfCAVEPreFrame();
            pfFrame();
            pfCAVEPostFrame();
            pfCAVEPreDrawFunc(NULL);
            world->frame();
            checkNetMessages();
            checkStdinMessages();
            if (CAVEgetbutton(CAVE_RKEY))
                world->reset();
            if (CAVEgetbutton(CAVE_GKEY))
                {
                ygPrintSceneGraph(world->root());
                printUsers(world);
                }
            pfSync();
            }
        }
    else
        {
        bool doSleep = false;
        if (getenv("YG_SERVER_DO_SLEEP"))
            doSleep = true;
        ygSetServerMode();
        while (!quitFlag)
            {
            checkNetMessages();
            checkStdinMessages();
            world->frame();
            if (doSleep)
                usleep(1000);
            }
        }
    world->unlinkLocalScene();
    delete world;
    ygHalt();
    pfCAVEHalt();
    pfuExitUtil();
    pfuExitInput();
    pfExit();
    return 0;
    }


static void initGraphics(pfChannel *chan,void *)
    {
    /* This is needed to keep it from crashing an Impact when the textures
      exceed texture memory */
    int i,size;
    static int first=1;
    if (first)
        {
        pfList *texList = pfuMakeSceneTexList(chan->getScene());
        if (getenv("YG_PF_SHOW_TEXTURES"))
            pfuDownloadTexList(texList, PFUTEX_SHOW);
        else
            pfuDownloadTexList(texList, PFUTEX_APPLY);
        for (i=0, size=0; i < texList->getNum(); i++)
            size += pfuGetTexSize((pfTexture *) texList->get(i));
        printf("TOTAL TEXTURE SIZE: %d bytes\n",size);
        first = 0;
        }
    }


static void checkStdinMessages(void)
    {
    fd_set readfds;
    struct timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    FD_ZERO(&readfds);
    FD_SET(0, &readfds);
    if (select(1,&readfds,NULL,NULL,&timeout))
        {
        if (FD_ISSET(0,&readfds))
            {
            char line[512];
            fgets(line,sizeof(line),stdin);
            if ((strlen(line)>0) && (line[strlen(line)-1] == '\n'))
                line[strlen(line)-1] = '\0';
            handleMessage(line);
            }
        }
    }


static void checkNetMessages(void)
    {
    static QUANTAnet_udp_c *sock = NULL;
    if (!sock)
        {
        int port=1967;
        sock = new QUANTAnet_udp_c;
        if (getenv("YG_MESSAGE_PORT"))
            port = atoi(getenv("YG_MESSAGE_PORT"));
        sock->init(port);
        sock->makeNonBlocking();
        }
    char message[1024];
    while (sock->receive(message,sizeof(message)) > 0)
        {
        printf("received message '%s'\n",message);
        handleMessage(message);
        }
    }


static bool allWhiteSpace(char *s)
    {
    while (*s)
        if (!isspace(*s))
            return false;
        else
            s++;
    return true;
    }


static void handleMessage(char *message)
    {
    if (allWhiteSpace(message))
        return;
    ygMessage msg;
    msg.parseString(message,NULL,NULL);
    if (msg.recipientName().length() > 0)
        ygWorld::world()->scheduleMessage(msg);
    else if (msg == "texcheck")
        {
        ygNode *node = ygWorld::world()->root();
        if (msg.numArgs() > 0)
            node = ygNodeDB::find(msg.stringArg(0),false);
        if (node)
            {
            int i,size=0;
            pfList * texList = pfuMakeTexList(node->pfnode());
            for (i=0, size=0; i < texList->getNum(); i++)
                size += pfuGetTexSize((pfTexture *) texList->get(i));
            printf("%s has %d textures, total size = %d bytes\n",
                node->name().c_str(),i,size);
            }
        }
    else if (msg == "stats")
        {
        ygNode *node = ygWorld::world()->root();
        if (msg.numArgs() > 0)
            node = ygNodeDB::find(msg.stringArg(0),false);
        if (node)
            node->printTimingStats();
        }
    else if (msg == "graph")
        {
        ygNode *node = ygWorld::world()->root();
        int depth = 99;
        if (msg.numArgs() > 0)
            {
            node = ygNodeDB::find(msg.stringArg(0),false);
        if (msg.numArgs() == 2)
            depth = msg.intArg(1);
            }
        if (node)
            ygPrintSceneGraph(node,depth);
        }
    else if (msg == "netstats")
        ygNet::showStats();
    else if (msg == "netreset")
        ygNet::reset();
    else if ((msg == "quit") || (msg == "exit"))
        quitFlag = true;
    else if (msg == "load")
        {
        if (msg.numArgs() == 1)
            ygWorld::world()->loadFromFile(msg.stringArg(0),ygWorld::world()->localRoot());
        else
            printf("ERROR: wrong number of arguments for load\n");
        }
    else if (msg == "create")
        {
        ygNode *node=NULL;
        if (msg.numArgs() == 1)
            {
            ygString name;
            ygNodeDB::createValidName(msg.stringArg(0),name);
            node = ygNodeDB::create(msg.stringArg(0),name);
            }
        else if (msg.numArgs() == 2)
            node = ygNodeDB::create(msg.stringArg(0),msg.stringArg(1));
        else
            printf("ERROR: wrong number of arguments for create()\n");
        if (node)
            printf("%s node '%s' created\n",node->className().c_str(),node->name().c_str());
        }
    else if (msg == "save")
        {
        if (msg.numArgs() == 1)
            ygWorld::world()->saveLocalScene(msg.stringArg(0));
        else if (msg.numArgs() == 2)
            {
            ygNode* node;
            node = ygNodeDB::find(msg.stringArg(1),false);
            if (node)
                ygWorld::world()->saveToFile(msg.stringArg(0),node);
            else
                msg.error("World","(node not found)");
            }
        else
            printf("ERROR: wrong number of arguments for save\n");
        }
    else if (msg == "delete")
        {
        if (msg.numArgs() == 1)
            {
            ygNode * node = ygNodeDB::find(msg.stringArg(0),false);
            if (node)
                ygWorld::world()->deleteTree(node);
            else
                msg.error("World","(node not found)");
            }
        else
            ygWorld::world()->deleteLocalScene();
        }
    else
        printf("WARNING: unrecognized command\n");
    }

#include "ygDSOLoader.h"
#include "ygCAVENavigator.h"
#include "ygCAVEHead.h"
#include "ygCAVETracker.h"
#include "ygCAVEWand.h"


static void addCAVEClasses(void)
    {
    ygDSOLoader::addConstructor("ygCAVENavigator",construct_ygCAVENavigator);
    ygDSOLoader::addConstructor("CAVENavigator",construct_ygCAVENavigator);
    ygDSOLoader::addConstructor("ygCAVEHead",construct_ygCAVEHead);
    ygDSOLoader::addConstructor("CAVEHead",construct_ygCAVEHead);
    ygDSOLoader::addConstructor("ygCAVETracker",construct_ygCAVETracker);
    ygDSOLoader::addConstructor("CAVETracker",construct_ygCAVETracker);
    ygDSOLoader::addConstructor("ygCAVEWand",construct_ygCAVEWand);
    ygDSOLoader::addConstructor("CAVEWand",construct_ygCAVEWand);
    }
