// Description: establishes a view point for the CAVE library rendering
//
//<b>notes:</b>
//<ul>
//<li> derived nodes inherit all messages from their base classes
//<li> creates a tracked transform from CAVE sensor 0
//<li> added automatically to scene at startup unless the --noview option is used
//</ul>
//
// Category: Foundation
// Author: Dave Pape
//           11/01/01
// Revision: 05/01/03 Alex Hill - added setUser() method to use another user perspective
//
#include <pfcave.h>
#include "ygWorld.h"
#include "ygCAVEViewer.h"
#include "ygNodeDB.h"

extern "C" ygNode* construct_ygCAVEViewer(const char* name,bool master) { return new ygCAVEViewer(name,master); }


static ygNavigator* findNavigator(ygNode * node)
    {
    if (node)
        {
        if (node->isOfClass("ygNavigator"))
            return (ygNavigator*) node;
        for (int i=0; i < node->numChildren(); i++)
            {
            ygNavigator *nav = findNavigator(node->child(i));
            if (nav)
                return nav;
            }
        }
    return NULL;
    }


ygCAVEViewer::ygCAVEViewer(const char* name,bool master) : ygTransform(name,master)
    {
    setClassName("ygCAVEViewer");
    navigator_ = NULL;
    user_ = NULL;
    CAVESetOption(CAVE_PROJ_INCLUDENAVIGATION,1);
    }


void ygCAVEViewer::reset(void)
    {
    navigator_ = findNavigator(ygWorld::world()->localRoot());
    ygTransform::reset();
    }


void ygCAVEViewer::app(void)
    {
    if (navigator_)
        {
        CAVENavLock();
        CAVENavLoadIdentity();
        CAVENavTranslate(navigator_->position()[0],
                         navigator_->position()[1],
                         navigator_->position()[2]);
        CAVENavRot(navigator_->orientation()[PF_Z],'z');
        CAVENavRot(navigator_->orientation()[PF_X],'x');
        CAVENavRot(navigator_->orientation()[PF_Y],'y');
        CAVENavUnlock();
        }
    if (user_ && user_->head())
        {
        pfMatrix mat;
        //get user navigation matrix
        user_->getMatrix(mat);
        pfCoord coord;
        mat.getOrthoCoord(&coord);
        //update navigation data
        CAVENavLock();
        CAVENavLoadIdentity();
        CAVENavTranslate(coord.xyz.vec[0],coord.xyz.vec[1],coord.xyz.vec[2]);
        CAVENavRot(coord.hpr.vec[0],'z');
        CAVENavRot(coord.hpr.vec[1],'x');
        CAVENavRot(coord.hpr.vec[2],'y');
        CAVENavUnlock();
        //get user head matrix
        ((ygTransform*)user_->head())->getMatrix(mat);
        mat.getOrthoCoord(&coord);
        //update head sensor
        CAVESensor1->x = coord.xyz.vec[0];
        CAVESensor1->y = coord.xyz.vec[1];
        CAVESensor1->z = coord.xyz.vec[2];
        CAVESensor1->azim = coord.hpr.vec[0];
        CAVESensor1->elev = coord.hpr.vec[1];
        CAVESensor1->roll = coord.hpr.vec[2];
        //get navigated head position
        user_->head()->getTransform(mat);
        mat.getOrthoCoord(&coord);
        setPosition(coord.xyz);
        setOrientation(coord.hpr.vec[1],coord.hpr.vec[2],coord.hpr.vec[0]);
        }
    else
        {
        dontSendNetMessage("m");
        pfVec3 pos, ori;
        //get CAVE head position
        CAVEGetPosition(CAVE_HEAD_NAV, pos.vec);
        //get CAVE head orientation
        CAVEGetOrientation(CAVE_HEAD_NAV, ori.vec);
        //set the position and orientation of the transform
        setPosition(pos);
        setOrientation(ori);
        }
    ygTransform::app();
    }


void ygCAVEViewer::setUser(const ygString& userName)
    {
    ygMessage msg;
    msg.parseString("updateInterval(10)",NULL,NULL);
    ygTransform::message(msg);
    user_ = (ygUser*)ygNodeDB::find(userName);    
    }


bool ygCAVEViewer::isCloned(ygNode* node)
    {
    if (node)
        {
        if (user_ == node)
            return true;
        else
            return false;
        }
    else
        {
        if (user_)
            return true;
        else
            return false;
        }
    }


