// Description: a base class for ygGeometry nodes
//
//<b>notes:</b>
//<ul>
//<li> all parent node messages are inherited
//<li> see reset method for default settings
//<li> the wall mode is used to turn collision detection on
//<li> the floor mode is used to specify whether the model is treated as terrain
//<li> the draw mode is set to false, ygGeometry is not draw but collision detection can still be enabled
//</ul>
//
// Category: Geometry
// Author: Alex Hill
//           11/01/02
// Revision: 09/01/04 Alex Hill - moved into the core executable and added "cull" message
//
#include <Performer/pf/pfGroup.h>
#include <Performer/pr/pfGeoSet.h>
#include <Performer/pf/pfGeode.h>
#include <ygWorld.h>
#include <ygNetMessages.h>
#include <ygPFObjectCache.h>
#include <Performer/pfdu.h>
#include "ygGeometry.h"

extern "C" ygNode* construct_ygGeometry(const char* name,bool master) { return new ygGeometry(name,master); }

ygGeometry::ygGeometry(const char* name,bool master) : ygNode(name,master)
    {
    setClassName("ygGeometry",true);
    isectMask = YG_ISECT_ALLMASK;
    floor = false;
    wall = false;
    draw = DRAW_ON;
    wireframe = false;
    cull = PFCF_BACK;
    pfnode()->setTravMask(PFTRAV_ISECT,isectMask,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
    pfnode()->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF,PF_SET);
    addNetMessage("floor",&floor,YG_NET_BOOL_NEGATE);
    addNetMessage("wall",&wall,YG_NET_BOOL_NEGATE);
    addNetMessage("draw",&draw,YG_NET_INT);
    addNetMessageMap("draw","off",DRAW_OFF);
    addNetMessageMap("draw","false",DRAW_OFF);
    addNetMessageMap("draw","true",DRAW_ON);
    addNetMessageMap("draw","on",DRAW_ON);
    addNetMessageMap("draw","local",DRAW_LOCAL);
    addNetMessageMap("draw","remote",DRAW_REMOTE);
    addNetMessage("wireframe",&wireframe,YG_NET_BOOL_NEGATE);
    addNetMessage("cull",&cull,YG_NET_INT);
    addNetMessageMap("cull","off",PFCF_OFF);
    addNetMessageMap("cull","none",PFCF_OFF);
    addNetMessageMap("cull","front",PFCF_FRONT);
    addNetMessageMap("cull","back",PFCF_BACK);
    addNetMessageMap("cull","both",PFCF_BOTH);
    }

ygGeometry::~ygGeometry(void)
    {
    }


void ygGeometry::reset(void)
    {
    //set floor to false
    resetNetMessage("floor");
    setFloor(floor);
    //set wall to false
    resetNetMessage("wall");
    setWall(wall);
    //set the draw flag to true
    resetNetMessage("draw");
    setDraw(draw);
    //set the wireframe flag to false
    resetNetMessage("wireframe");
    if (wireframe)
        setWireframe(wireframe);
    //set the cull mode to back faces
    resetNetMessage("cull");
    if (cull != PFCF_BACK)
        setCull(cull);
    ygNode::reset();
    }


void ygGeometry::message(const ygMessage& msg)
    {
    //set collision detection status
    if (msg == "wall")
        {
        setWall(wall);
        }
    //set walk navigation status
    else if (msg == "floor")
        {
        setFloor(floor);
        }
    //set if the object is draw
    else if (msg == "draw")
        {
        setDraw(draw);
        }
    //set wireframe mode
    else if (msg == "wireframe")
        {
        setWireframe(wireframe);
        }
    //set cull mode
    else if (msg == "cull")
        {
        setCull(cull);
        }
    else
        ygNode::message(msg);
    }


void ygGeometry::app(void)
    {
    if (isChanged(YG_CHANGE_GEOMETRY))
        {
        setFloor(floor);
        setWall(wall);
        }
    ygNode::app();
    }


void ygGeometry::setWall(bool val)
    {
    wall = val;
    if (val)
        isectMask |= YG_ISECT_WALLMASK;
    else
        isectMask &= ~YG_ISECT_WALLMASK;
    pfnode()->setTravMask(PFTRAV_ISECT,isectMask,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
    }


void ygGeometry::setFloor(bool val)
    {
    floor = val;
    if (val)
        isectMask |= YG_ISECT_FLOORMASK;
    else
        isectMask &= ~YG_ISECT_FLOORMASK;
    pfnode()->setTravMask(PFTRAV_ISECT,isectMask,PFTRAV_SELF|PFTRAV_DESCEND,PF_SET);
    }


void ygGeometry::setDraw(int val)
    {
    draw = val;
    if (val == DRAW_ON || (isNetMaster() && val == DRAW_LOCAL) || (!isNetMaster() && val == DRAW_REMOTE))
        pfnode()->setTravMask(PFTRAV_DRAW,0xffffffff,PFTRAV_SELF,PF_SET);
    else
        pfnode()->setTravMask(PFTRAV_DRAW,0,PFTRAV_SELF,PF_SET);
    }


void ygGeometry::setWireframe(bool val)
    {
    wireframe = val;
    if (val)
        setGeoState(PFSTATE_ENWIREFRAME,PF_ON,pfnode());
    else
        setGeoState(PFSTATE_ENWIREFRAME,PF_OFF,pfnode());
    }


void ygGeometry::setCull(int mode)
    {
    cull = mode;
    setGeoState(PFSTATE_CULLFACE,cull,pfnode());
    }


void ygGeometry::setGeoState(uint64_t mode, int state, pfNode* node)
    {
    int i;
    //if node type is pfGeode then set geoset properties
    if (node->isOfType(pfGeode::getClassType()))
        {
        pfGeode *geode = (pfGeode *)node;
        //for each geode, call setGeosetTexture
        for (i=0; i < geode->getNumGSets(); i++)
            {
            pfGeoSet* gset = geode->getGSet(i);
            gset->getGState()->setMode(mode,state);
            }
        }
    //else, if node type is pfGroup
    else if (node->isOfType(pfGroup::getClassType()))
        {
        pfGroup *group = (pfGroup *)node;
        //call this function recursively on all children
        for (i=0; i < group->getNumChildren(); i++)
            setGeoState(mode,state,group->getChild(i));
        }
    }
